/*******************************************************************************
 * I9 TECHNOLOGIES CONFIDENTIAL
 * __________________
 * Copyright © 2013 i9 Technologies 
 * All Rights Reserved.
 * 
 * NOTICE:  All information contained herein is, and remains the property 
 * of i9 Technologies and its suppliers, if any.  The intellectual and technical 
 * concepts contained herein are proprietary to i9 Technologies and its 
 * suppliers and may be covered by U.S. and Foreign Patents, patents in 
 * process, and are protected by trade secret or copyright law. Dissemination 
 * of this information or reproduction of this material is strictly forbidden 
 * unless prior written permission is obtained from i9 Technologies.
 ******************************************************************************/
package com.xmlnamespace.panel.vaadin.ui;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import com.vaadin.event.DataBoundTransferable;
import com.vaadin.event.MouseEvents.ClickEvent;
import com.vaadin.event.MouseEvents.ClickListener;
import com.vaadin.event.Transferable;
import com.vaadin.event.TransferableImpl;
import com.vaadin.event.dd.DragAndDropEvent;
import com.vaadin.event.dd.DragSource;
import com.vaadin.event.dd.DropHandler;
import com.vaadin.event.dd.DropTarget;
import com.vaadin.event.dd.TargetDetails;
import com.vaadin.event.dd.TargetDetailsImpl;
import com.vaadin.event.dd.acceptcriteria.AcceptCriterion;
import com.vaadin.event.dd.acceptcriteria.ServerSideCriterion;
import com.vaadin.server.PaintException;
import com.vaadin.server.PaintTarget;
import com.vaadin.server.Resource;
import com.vaadin.server.ResourceReference;
import com.vaadin.shared.MouseEventDetails;
import com.vaadin.shared.communication.URLReference;
import com.vaadin.ui.AbstractComponent;
import com.vaadin.ui.Component;
import com.vaadin.ui.LegacyComponent;
import com.vaadin.util.ReflectTools;
import com.xmlnamespace.panel.vaadin.shared.ui.UserAccountPopupItem;
import com.xmlnamespace.panel.vaadin.shared.ui.UserButtonItem;
import com.xmlnamespace.panel.vaadin.shared.ui.UserComponentClientRpc;
import com.xmlnamespace.panel.vaadin.shared.ui.UserComponentItem;
import com.xmlnamespace.panel.vaadin.shared.ui.UserComponentItem.UserComponentItemChangeListener;
import com.xmlnamespace.panel.vaadin.shared.ui.UserComponentItem.UserComponentItemChangedEvent;
import com.xmlnamespace.panel.vaadin.shared.ui.UserComponentServerRpc;
import com.xmlnamespace.panel.vaadin.shared.ui.UserComponentState;

/**
 * Component used to display an iSymphony user, its current conversation information
 * and accounts.
 * 
 * Provides listeners to detect click events on the status icon, user buttons, 
 * conversation switcher buttons and account items in the account popup. Also provides
 * double click listeners on the entire component.
 * 
 * Provides drag and drop functionality that allows accepted drop classes to be specified
 * and allows users to specify the object that are dragged for both the user and the current
 * conversation.
 */
public class UserComponent extends AbstractComponent implements
        UserComponentItemChangeListener, DragSource, DropTarget, LegacyComponent {
    
    /** The user account popup item map. */
    private Map<Double, UserAccountPopupItem> userAccountPopupItemMap = new HashMap<Double, UserAccountPopupItem>();
    
    /** The user button item map. */
    private Map<Double, UserButtonItem> userButtonItemMap = new HashMap<Double, UserButtonItem>();

    /** The accepted drop classes. */
    private Class<?>[] acceptedDropClasses;
    
    /** The drop handler. */
    private UserComponentDropHandler dropHandler = new UserComponentDropHandler();
    
    /** The user drag object. */
    private Object userDragObject;
    
    /** The conv drag object. */
    private Object convDragObject;
    
    private long convStartTime = -1;
    
    /** The rpc. */
    private UserComponentServerRpc rpc = new UserComponentServerRpc() {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;

        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.shared.ui.UserComponentServerRpc#userAccountItemClicked(double, com.vaadin.shared.MouseEventDetails)
         */
        @Override
        public void userAccountItemClicked(double itemId,
                MouseEventDetails details) {
            
            //If not licensed do nothing
            if(!getState().licensed) {
                return;
            }
            
            //Locate the server side item object
            UserAccountPopupItem item = userAccountPopupItemMap.get(itemId);
            
            //Fire event
            fireEvent(new UserComponent.UserAccountPopupItemClickEvent(UserComponent.this, item, details));
        }

        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.shared.ui.UserComponentServerRpc#userButtonItemClicked(double, com.vaadin.shared.MouseEventDetails)
         */
        @Override
        public void userButtonItemClicked(double itemId,
                MouseEventDetails details) {
            
            //If not licensed do nothing
            if(!getState().licensed) {
                return;
            }
            
            //Locate server side item object
            UserButtonItem item = userButtonItemMap.get(itemId);
            
            //Fire event
            fireEvent(new UserComponent.UserButtonItemClickEvent(UserComponent.this, item, details));
        }

        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.shared.ui.UserComponentServerRpc#statusIconClicked(com.vaadin.shared.MouseEventDetails)
         */
        @Override
        public void statusIconClicked(MouseEventDetails details) {
            
            //If not licensed do nothing
            if(!getState().licensed) {
                return;
            }
            
            fireEvent(new UserComponent.StatusIconClickEvent(UserComponent.this, details));
        }

        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.shared.ui.UserComponentServerRpc#convSwitcherUpClicked(com.vaadin.shared.MouseEventDetails)
         */
        @Override
        public void convSwitcherUpClicked(MouseEventDetails details) {
            
            //If not licensed do nothing
            if(!getState().licensed) {
                return;
            }
            
            fireEvent(new UserComponent.ConvSwitcherClickEvent(UserComponent.this, details, true));
        }

        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.shared.ui.UserComponentServerRpc#convSwitcherDownClicked(com.vaadin.shared.MouseEventDetails)
         */
        @Override
        public void convSwitcherDownClicked(MouseEventDetails details) {
            
            //If not licensed do nothing
            if(!getState().licensed) {
                return;
            }
            
            fireEvent(new UserComponent.ConvSwitcherClickEvent(UserComponent.this, details, false));
        }

        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.shared.ui.UserComponentServerRpc#componentDoubleClicked(com.vaadin.shared.MouseEventDetails)
         */
        @Override
        public void componentDoubleClicked(MouseEventDetails details) {
            
            //If not licensed do nothing
            if(!getState().licensed) {
                return;
            }
            
            fireEvent(new UserComponent.UserComponentDoubleClickEvent(UserComponent.this, details));
        }
        
        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.shared.ui.UserComponentServerRpc#updateConvTime()
         */
        @Override
        public void updateConvTime() {
            setConvStartTime(convStartTime);
        }
        
    };
    
    /**
     * Instantiates a new user component.
     */
    public UserComponent() {
        registerRpc(rpc);
    }
    
    /* (non-Javadoc)
     * @see com.vaadin.ui.AbstractComponent#attach()
     */
    @Override
    public void attach() {
        super.attach();
        
        // Update the conversation start time with the stored value, to handle page refreshes & detach/re-attach scenarios
        setConvStartTime(convStartTime);
    }
    
    /**
     * Sets the licensed flag.
     *
     * @param licensed if false only the username and unlicensed text will show
     * in the component and all actions will be disabled
     */
    public void setLicensed(boolean licensed) {
        getState().licensed = licensed;
        markAsDirty();
    }
    
    /**
     * Checks if this component is licensed.
     *
     * @return true, if this component is license false otherwise
     */
    public boolean isLicensed() {
        return getState().licensed;
    }
    
    /**
     * Sets the unlicensed text witch will 
     * be displayed in the component if 
     * setLicensed(false) is called.
     *
     * @param unlicensedText the new unlicensed text
     */
    public void setUnlicensedText(String unlicensedText) {
        getState().unlicensedText = unlicensedText;
        markAsDirty();
    }
    
    /**
     * Gets the unlicensed text.
     *
     * @return the unlicensed text
     */
    public String getUnlicensedText() {
        return getState().unlicensedText;
    }
    
    /**
     * Sets the status icon.
     *
     * @param icon the new status icon
     */
    public void setStatusIcon(Resource icon) {
        if (icon == null) {
            getState().statusIcon = null;
        } else {
            getState().statusIcon = new ResourceReference(icon, this,
                    "status-icon");
        }
        markAsDirty();
    }

    /**
     * Gets the status icon.
     *
     * @return the status icon
     */
    public Resource getStatusIcon() {
        URLReference ref = getState().statusIcon;
        if (ref == null) {
            return null;
        }

        return ((ResourceReference) ref).getResource();
    }

    /**
     * Sets the user status icon tool tip.
     *
     * @param toolTip the new user status icon tool tip
     */
    public void setUserStatusIconToolTip(String toolTip) {
        getState().statusIconToolTip = toolTip;
        markAsDirty();
    }
    
    /**
     * Gets the user status icon tool tip.
     *
     * @return the user status icon tool tip
     */
    public String getUserStatusIconToolTip() {
        return getState().statusIconToolTip;
    }
    /**
     * Sets the username.
     *
     * @param username the new username
     */
    public void setUsername(String username) {
        getState().username = username;
        markAsDirty();
    }

    /**
     * Gets the username.
     *
     * @return the username
     */
    public String getUsername() {
        return getState().username;
    }

    /**
     * Sets the status.
     *
     * @param status the new status
     */
    public void setStatus(String status) {
        getState().status = status;
        markAsDirty();
    }

    /**
     * Gets the status.
     *
     * @return the status
     */
    public String getStatus() {
        return getState().status;
    }

    /**
     * Sets the conv count.
     *
     * @param convCount the new conv count
     */
    public void setConvCount(String convCount) {
        getState().convCount = convCount;
        markAsDirty();
    }

    /**
     * Gets the conv count.
     *
     * @return the conv count
     */
    public String getConvCount() {
        return getState().convCount;
    }

    /**
     * Sets the conv name.
     *
     * @param convName the new conv name
     */
    public void setConvName(String convName) {
        getState().convName = convName;
        markAsDirty();
    }

    /**
     * Gets the conv name.
     *
     * @return the conv name
     */
    public String getConvName() {
        return getState().convName;
    }

    /**
     * Sets if the conv switcher is visible or not.
     *
     * @param convSwitcherVisible true for the switcher to be visible; false otherwise
     */
    public void setConvSwitcherVisible(boolean convSwitcherVisible) {
        getState().convSwitcherVisible = convSwitcherVisible;
        markAsDirty();
    }

    /**
     * Checks if is conv switcher is visible.
     *
     * @return true, if is conv switcher is visible
     */
    public boolean isConvSwitcherVisible() {
        return getState().convSwitcherVisible;
    }

    /**
     * Sets the conv start time in seconds. If the value is > 0 the
     * timer will be shown and will start counting at the given number
     * of seconds.
     * 
     * If -1 is given the timer will be hidden.
     *
     * @param seconds the number of seconds to start the timer at or -1 to hide the timer
     */
    public void setConvStartTime(long startTime) {
        this.convStartTime = startTime;
        if (startTime < 0)
            getRpcProxy(UserComponentClientRpc.class).setConvCurrentTime(startTime);
        else
            getRpcProxy(UserComponentClientRpc.class).setConvCurrentTime(new Date().getTime() - startTime);
    }
    
    /**
     * Gets the conversation styles.
     *
     * @return the conversation styles
     */
    public Set<String> getConvStyles() {
        return getState().convStyles;
    }
    
    /**
     * Adds a conversation style.
     * 
     * This is the style that will be applied to the 
     * conversation count, conversation name, and 
     * conversation time elements.
     *
     * @param style the style to add
     */
    public void addConvStyle(String style) {
        getState().convStyles.add(style);
        markAsDirty();
    }
    
    /**
     * Removes a conversation style.
     *
     * @param style the style to remove
     */
    public void removeConvStyle(String style) {
        getState().convStyles.remove(style);
        markAsDirty();
    }
    
    /**
     * Gets the user drag object.
     *
     * @return the user drag object
     */
    public Object getUserDragObject() {
        return userDragObject;
    }

    /**
     * Sets the user drag object.
     * 
     * This is the object that will be dragged if one of the following is true
     * 
     * - The user starts the drag from one of the header elements(status icon, username, status)
     * - The conversation drag object is null and the user drags from any point in the component
     *
     * @param userDragObject the new user drag object or null for no user drag
     */
    public void setUserDragObject(Object userDragObject) {
        this.userDragObject = userDragObject;
        getState().hasUserDragObject = userDragObject != null;
        markAsDirty();
    }

    /**
     * Gets the conv drag object.
     *
     * @return the conv drag object
     */
    public Object getConvDragObject() {
        return convDragObject;
    }

    /**
     * Sets the conversation drag object.
     * 
     * This is the object that will be dragged if one of the following is true
     * 
     * - User drags from the conversation name element
     * - User drag object is null and the user drags from any point in the component
     *
     * @param convDragObject the new conv drag object or null for no conversation drag
     */
    public void setConvDragObject(Object convDragObject) {
        this.convDragObject = convDragObject;
        getState().hasConvDragObject = convDragObject != null;
        markAsDirty();
    }

    /**
     * Adds a user account popup item with the given name and type.
     *
     * @param name the name of the account
     * @param type the type of the account
     * @return the user account popup item created
     */
    public UserAccountPopupItem addUserAccountPopupItem(String name, String type) {

        // Create item and add to the registry
        UserAccountPopupItem item = new UserAccountPopupItem(name, type);
        getState().userAccountPopupItems.add(item);
        userAccountPopupItemMap.put(item.getId(), item);

        // Inform client of addition
        getRpcProxy(UserComponentClientRpc.class).addAccountPopupItem(item);

        // Add listener to detect changes on the item
        item.addChangeListener(this);

        return item;
    }

    /**
     * Removes a user account popup item.
     *
     * @param item the item to remove
     */
    public void removeUserAccountPopupItem(UserAccountPopupItem item) {

        // Remove item change listener
        item.removeChangeListener(this);

        // Inform the client of the removal
        getRpcProxy(UserComponentClientRpc.class).removeAccountPopupItem(item);

        // Remove item from the registry
        getState().userAccountPopupItems.remove(item);
        userAccountPopupItemMap.remove(item.getId());
    }

    /**
     * Adds a user button item with the give icon.
     *
     * @param icon the icon that represents the button
     * @return the user button item created
     */
    public UserButtonItem addUserButtonItem(Resource icon) {

        // Determine the url reference for the button icon
        URLReference iconRef = null;
        if (icon != null) {
            iconRef = new ResourceReference(icon, this, "button-icon");
        }

        // Create item and add to the registry
        UserButtonItem item = new UserButtonItem(iconRef);
        getState().userButtonItems.add(item);
        userButtonItemMap.put(item.getId(), item);

        // Inform client of addition
        getRpcProxy(UserComponentClientRpc.class).addUserButtonItem(item);

        // Add listener to detect changes on the item
        item.addChangeListener(this);

        return item;
    }

    /**
     * Removes a user button item.
     *
     * @param item the item to remove
     */
    public void removeUserButtonItem(UserButtonItem item) {

        // Remove item change listener
        item.removeChangeListener(this);

        // Inform the client of the removal
        getRpcProxy(UserComponentClientRpc.class).removeUserButtonItem(item);

        // Remove item from the registry
        getState().userButtonItems.remove(item);
        userButtonItemMap.remove(item.getId());
    }

    /**
     * Convenience method used to set a new icon on a button item.
     * 
     * Users can also set a URLReference on the item itself
     *
     * @param item the item to set the icon on
     * @param icon the new icon
     */
    public void setButtonItemIcon(UserButtonItem item, Resource icon) {
        item.setIcon(new ResourceReference(icon, this, "button-icon"));
    }

    /**
     * Gets the accepted drop classes.
     *
     * @return the accepted drop classes
     */
    public Class<?>[] getAcceptedDropClasses() {
        return acceptedDropClasses;
    }

    /**
     * Sets the accepted drop classes that are allowed to be dropped on this component.
     *
     * @param acceptedDropClasses the new accepted drop classes
     */
    public void setAcceptedDropClasses(Class<?>[] acceptedDropClasses) {
        this.acceptedDropClasses = acceptedDropClasses;
    }

    /**
     * Sets the drag mode.
     * 
     * UserComponent.DragMode.NONE will disable all drags from the component
     * UserComponent.DragMode.DRAG will enable drags from the component
     *
     * @param dragMode the new drag mode
     */
    public void setDragMode(UserComponent.DragMode dragMode) {
        getState().dragMode = dragMode.ordinal();
        markAsDirty();
    }
    
    /**
     * Gets the drag mode.
     *
     * @return the drag mode
     */
    public UserComponent.DragMode getDragMode() {
        return UserComponent.DragMode.values()[getState().dragMode];
    }
    
    /**
     * Adds a listener to detect when the component is double clicked
     *
     * @param listener the listener
     */
    public void addUserComponentDoubleClickListener(UserComponentDoubleClickListener listener) {
        addListener(UserComponentDoubleClickEvent.class, listener,
                UserComponentDoubleClickListener.clickMethod);
    }
    
    /**
     * Removes a user component double click listener.
     *
     * @param listener the listener
     */
    public void removeUserComponentDoubleClickListener(UserComponentDoubleClickListener listener) {
        removeListener(UserComponentDoubleClickEvent.class, listener,
                UserComponentDoubleClickListener.clickMethod);
    }
    
    /**
     * Adds a listener to detect when a user account popup item is clicked
     *
     * @param listener the listener
     */
    public void addUserAccountPopupItemClickListener(UserAccountPopupItemClickListener listener) {
        addListener(UserAccountPopupItemClickEvent.class, listener,
                UserAccountPopupItemClickListener.clickMethod);
    }

    /**
     * Removes a user account popup item click listener.
     *
     * @param listener the listener
     */
    public void removeUserAccountPopupItemClickListener(UserAccountPopupItemClickListener listener) {
        removeListener(UserAccountPopupItemClickEvent.class, listener,
                UserAccountPopupItemClickListener.clickMethod);
    }
    
    /**
     * Adds a listener to detect when a user button item is clicked
     *
     * @param listener the listener
     */
    public void addUserButtonItemClickListener(UserButtonItemClickListener listener) {
        addListener(UserButtonItemClickEvent.class, listener,
                UserButtonItemClickListener.clickMethod);
    }
    
    /**
     * Removes a user button item click listener.
     *
     * @param listener the listener
     */
    public void removeUserButtonItemClickListener(UserButtonItemClickListener listener) {
        removeListener(UserButtonItemClickEvent.class, listener,
                UserButtonItemClickListener.clickMethod);
    }
    
    /**
     * Adds a listener to detect when the status icon is clicked
     *
     * @param listener the listener
     */
    public void addStatusIconClickListener(StatusIconClickListener listener) {
        addListener(StatusIconClickEvent.class, listener,
                StatusIconClickListener.clickMethod);
    }
    
    /**
     * Removes a status icon click listener.
     *
     * @param listener the listener
     */
    public void removeStatusIconClickListener(StatusIconClickListener listener) {
        removeListener(StatusIconClickEvent.class, listener,
                StatusIconClickListener.clickMethod);
    }
    
    /**
     * Adds a listener to determine when the up or down button on the conversation switcher is clicked.
     *
     * @param listener the listener
     */
    public void addConvSwitcherClickListener(ConvSwitcherClickListener listener) {
        addListener(ConvSwitcherClickEvent.class, listener,
                ConvSwitcherClickListener.clickMethod);
    }
    
    /**
     * Removes a conversation switcher click listener.
     *
     * @param listener the listener
     */
    public void removeConvSwitcherClickListener(ConvSwitcherClickListener listener) {
        removeListener(ConvSwitcherClickEvent.class, listener,
                ConvSwitcherClickListener.clickMethod);
    }
    
    /**
     * Adds a listener to detect when objects are dropped on this component
     * 
     * Use in conjunction with setAcceptedDropClasses()
     *
     * @param listener the listener
     */
    public void addObjectDropListener(ObjectDropListener listener) {
        addListener(ObjectDropEvent.class, listener,
                ObjectDropListener.DROP_METHOD);
    }
    
    /**
     * Removes an object drop listener.
     *
     * @param listener the listener
     */
    public void removeObjectDropListener(ObjectDropListener listener) {
        removeListener(ObjectDropEvent.class, listener,
                ObjectDropListener.DROP_METHOD);
    }
    
    /**
     * Gets the data object from a drag and drop event.
     *
     * @param event the event
     * @return the data object from the drag and drop event
     */
    private static Object getDataFromDragAndDropEvent(DragAndDropEvent event) {
        
        //If drop is a DragAndDropComponent grab it from the custom method else if it is a DataBoundTransferable use the get item id class
        if(event.getTransferable().getSourceComponent() instanceof DragAndDropComponent) {
            return ((DragAndDropComponent)event.getTransferable().getSourceComponent()).getDragObject();
        } else if(event.getTransferable() instanceof DataBoundTransferable) {
            return ((DataBoundTransferable)event.getTransferable()).getItemId();
        } else if(event.getTransferable() instanceof UserComponentTransferable) {
            return ((UserComponentTransferable)event.getTransferable()).getObject();
        } else if(event.getTransferable() instanceof CallComponent.CallComponentTransferable) {
            return ((CallComponent.CallComponentTransferable)event.getTransferable()).getObject();
        } else {
            throw new IllegalArgumentException("Drop type not known");
        }       
    }
    
    /* (non-Javadoc)
     * @see com.vaadin.ui.AbstractComponent#getState()
     */
    @Override
    protected UserComponentState getState() {
        return (UserComponentState) super.getState();
    }

    /* (non-Javadoc)
     * @see com.vaadin.event.dd.DropTarget#getDropHandler()
     */
    @Override
    public DropHandler getDropHandler() {
        return dropHandler;
    }

    /* (non-Javadoc)
     * @see com.vaadin.event.dd.DropTarget#translateDropTargetDetails(java.util.Map)
     */
    @Override
    public TargetDetails translateDropTargetDetails(
            Map<String, Object> clientVariables) {
        return new TargetDetailsImpl(clientVariables, this);
    }

    /* (non-Javadoc)
     * @see com.vaadin.event.dd.DragSource#getTransferable(java.util.Map)
     */
    @Override
    public Transferable getTransferable(Map<String, Object> rawVariables) {
        return new UserComponentTransferable(this, rawVariables);
    }

    /*
     * (non-Javadoc)
     * 
     * @see com.xmlnamespace.panel.vaadin.shared.ui.UserComponentItem.
     * UserComponentItemChangeListener
     * #userComponentItemChanged(com.xmlnamespace.
     * panel.vaadin.shared.ui.UserComponentItem.UserComponentItemChangedEvent)
     */
    @Override
    public void userComponentItemChanged(UserComponentItemChangedEvent event) {

        UserComponentItem item = event.getItem();

        // Inform client of item update
        if (item instanceof UserAccountPopupItem) {
            getRpcProxy(UserComponentClientRpc.class).updateAccountPopupItem(
                    (UserAccountPopupItem) item);
        } else if (item instanceof UserButtonItem) {
            getRpcProxy(UserComponentClientRpc.class).updateUserButtonItem(
                    (UserButtonItem) item);
        }
    }
    
    /* (non-Javadoc)
     * @see com.vaadin.server.VariableOwner#changeVariables(java.lang.Object, java.util.Map)
     */
    @Override
    public void changeVariables(Object source, Map<String, Object> variables) {}
    
    /* (non-Javadoc)
     * @see com.vaadin.ui.LegacyComponent#paintContent(com.vaadin.server.PaintTarget)
     */
    @Override
    public void paintContent(PaintTarget target) throws PaintException {
        
        //Only add the drop handler if there are accepted drop classes
        if (acceptedDropClasses != null && acceptedDropClasses.length != 0) {
            dropHandler.getAcceptCriterion().paint(target);
        }
    }
        
    /**
     * Event fired when a user double clicks the user component
     */
    public static class UserComponentDoubleClickEvent extends ClickEvent {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;

        /**
         * Instantiates a new user component double click event.
         *
         * @param source the source
         * @param details the details
         */
        public UserComponentDoubleClickEvent(Component source,
                MouseEventDetails details) {
            super(source, details);
        }
        
    }
    
    /**
     * Base class for user item click events
     */
    public static class UserComponentItemClickEvent extends ClickEvent {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;

        /** The item. */
        private UserComponentItem item;

        /**
         * Instantiates a new user component item click event.
         *
         * @param source the source
         * @param item the item
         * @param details the details
         */
        public UserComponentItemClickEvent(Component source, UserComponentItem item,
                MouseEventDetails details) {
            super(source, details);
            this.item = item;
        }

        /**
         * Gets the item that was clicked on.
         *
         * @return the item that was clicked on
         */
        public UserComponentItem getItem() {
            return item;
        }
        
    }
    
    /**
     * Event fired when a user clicks an account popup item
     */
    public static class UserAccountPopupItemClickEvent extends UserComponentItemClickEvent {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;

        /**
         * Instantiates a new user account popup item click event.
         *
         * @param source the source
         * @param item the item
         * @param details the details
         */
        public UserAccountPopupItemClickEvent(Component source,
                UserAccountPopupItem item, MouseEventDetails details) {
            super(source, item, details);
        }

        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.ui.UserComponent.UserComponentItemClickEvent#getItem()
         */
        @Override
        public UserAccountPopupItem getItem() {
            return (UserAccountPopupItem)super.getItem();
        }
        
    }
    
    /**
     * Event fired when a user clicks a button item
     */
    public static class UserButtonItemClickEvent extends UserComponentItemClickEvent {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;

        /**
         * Instantiates a new user button item click event.
         *
         * @param source the source
         * @param item the item
         * @param details the details
         */
        public UserButtonItemClickEvent(Component source,
                UserButtonItem item, MouseEventDetails details) {
            super(source, item, details);
        }
        
        /* (non-Javadoc)
         * @see com.xmlnamespace.panel.vaadin.ui.UserComponent.UserComponentItemClickEvent#getItem()
         */
        @Override
        public UserButtonItem getItem() {
            return (UserButtonItem)super.getItem();
        }

    }
    
    /**
     * Event fired when a user clicks on the status icon
     */
    public static class StatusIconClickEvent extends ClickEvent {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;

        /**
         * Instantiates a new status icon click event.
         *
         * @param source the source
         * @param details the details
         */
        public StatusIconClickEvent(Component source, MouseEventDetails details) {
            super(source, details);
        }
        
    }
    
    /**
     * Event fired when a user clicks one of the conversation switcher buttons
     */
    public static class ConvSwitcherClickEvent extends ClickEvent {
        
        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;
        
        /** The up button. */
        private boolean upButton;
        
        /**
         * Instantiates a new conv switcher click event.
         *
         * @param source the source
         * @param details the details
         * @param upButton the up button
         */
        public ConvSwitcherClickEvent(Component source, MouseEventDetails details, boolean upButton) {
            super(source, details);
            this.upButton = upButton;
        }
        
        /**
         * Checks if the up button was clicked.
         *
         * @return true, if the up button was clicked; false if the down button was clicked
         */
        public boolean isUpButton() {
            return upButton;
        }
        
    }
    
    /**
     * Event fired when an object is successfully dropped on the user component
     */
    public static class ObjectDropEvent extends Component.Event {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;
        
        /** The event. */
        private DragAndDropEvent event;
        
        /** The drop object. */
        private Object dropObject;
        
        /** The drop item. */
        private UserComponentItem dropItem;

        /**
         * Instantiates a new object drop event.
         *
         * @param source the source
         * @param dropObject the drop object
         * @param dropItem the drop item
         */
        public ObjectDropEvent(DragAndDropEvent event, Component source, Object dropObject, UserComponentItem dropItem) {
            super(source);
            this.event= event;
            this.dropObject = dropObject;
            this.dropItem = dropItem;
        }

        /**
         * Gets the object that was dropped.
         *
         * @return the object that was dropped
         */
        public Object getDropObject() {
            return dropObject;
        }

        /**
         * Gets the user component item that was dropped on
         *
         * @return the user component item that was dropped on or null if 
         * the drop did not occur on an user component item
         */
        public UserComponentItem getDropItem() {
            return dropItem;
        }
        
        /**
         * Gets the DragAndDropEvent that caused this event.
         *
         * @return the drag and drop event
         */
        public DragAndDropEvent getEvent() {
            return event;
        }
        
    }
    
    /**
     * The listener interface for receiving userComponentDoubleClick events.
     * The class that is interested in processing a userComponentDoubleClick
     * event implements this interface, and the object created
     * with that class is registered with a component using the
     * component's <code>addUserComponentDoubleClickListener<code> method. When
     * the userComponentDoubleClick event occurs, that object's appropriate
     * method is invoked.
     *
     * @see UserComponentDoubleClickEvent
     */
    public interface UserComponentDoubleClickListener extends ClickListener {
    }
    
    /**
     * The listener interface for receiving userAccountPopupItemClick events.
     * The class that is interested in processing a userAccountPopupItemClick
     * event implements this interface, and the object created
     * with that class is registered with a component using the
     * component's <code>addUserAccountPopupItemClickListener<code> method. When
     * the userAccountPopupItemClick event occurs, that object's appropriate
     * method is invoked.
     *
     * @see UserAccountPopupItemClickEvent
     */
    public interface UserAccountPopupItemClickListener extends ClickListener {
    }
    
    /**
     * The listener interface for receiving userButtonItemClick events.
     * The class that is interested in processing a userButtonItemClick
     * event implements this interface, and the object created
     * with that class is registered with a component using the
     * component's <code>addUserButtonItemClickListener<code> method. When
     * the userButtonItemClick event occurs, that object's appropriate
     * method is invoked.
     *
     * @see UserButtonItemClickEvent
     */
    public interface UserButtonItemClickListener extends ClickListener {
    }
    
    /**
     * The listener interface for receiving statusIconClick events.
     * The class that is interested in processing a statusIconClick
     * event implements this interface, and the object created
     * with that class is registered with a component using the
     * component's <code>addStatusIconClickListener<code> method. When
     * the statusIconClick event occurs, that object's appropriate
     * method is invoked.
     *
     * @see StatusIconClickEvent
     */
    public interface StatusIconClickListener extends ClickListener {
    }
    
    /**
     * The listener interface for receiving convSwitcherClick events.
     * The class that is interested in processing a convSwitcherClick
     * event implements this interface, and the object created
     * with that class is registered with a component using the
     * component's <code>addConvSwitcherClickListener<code> method. When
     * the convSwitcherClick event occurs, that object's appropriate
     * method is invoked.
     *
     * @see ConvSwitcherClickEvent
     */
    public interface ConvSwitcherClickListener extends ClickListener {  
    }
    
    /**
     * The listener interface for receiving objectDrop events.
     * The class that is interested in processing a objectDrop
     * event implements this interface, and the object created
     * with that class is registered with a component using the
     * component's <code>addObjectDropListener<code> method. When
     * the objectDrop event occurs, that object's appropriate
     * method is invoked.
     *
     * @see ObjectDropEvent
     */
    public interface ObjectDropListener extends Serializable {

        /** The Constant DROP_METHOD. */
        public static final Method DROP_METHOD = ReflectTools
                .findMethod(ObjectDropListener.class, "onDrop",
                        ObjectDropEvent.class);

        /**
         * On drop.
         *
         * @param event the event
         */
        public void onDrop(ObjectDropEvent event);
        
    }
    
    /**
     * Drag modes for the user component.
     */
    public enum DragMode {
        
        /** Disables all drag functionality. */
        NONE,
        
        /** Enables drag functionality. */
        DRAG
    }
    
    /**
     * Transferable used for drags from the user component
     */
    public class UserComponentTransferable extends TransferableImpl {

        /** The object. */
        private Object object;
        
        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;
        
        /** The Constant TYPE_ID. */
        private static final String TYPE_ID = "object-type";
        
        /** The Constant USER_TYPE. */
        private static final String USER_TYPE = "user";
        
        /** The Constant CONV_TYPE. */
        private static final String CONV_TYPE = "conv";

        /**
         * Instantiates a new user component transferable.
         *
         * @param sourceComponent the source component
         * @param rawVariables the raw variables
         */
        public UserComponentTransferable(Component sourceComponent,
                Map<String, Object> rawVariables) {
            super(sourceComponent, rawVariables);
            
            String type = (String)getData(TYPE_ID);
            if(type != null) {
                if(type.equals(USER_TYPE)) {
                    object = userDragObject;
                } else if (type.equals(CONV_TYPE)){
                    object = convDragObject;
                }
            }
        }
        
        /**
         * Gets the drag object.
         *
         * @return the drag object
         */
        public Object getObject() {
            return object;
        }
        
    }
    
    /**
     * Handler used to detect drops on the user component
     */
    private class UserComponentDropHandler implements DropHandler {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;
        
        /** The Constant DROP_ITEM_DETAIL_ID. */
        private static final String DROP_ITEM_DETAIL_ID = "item-id-dropped-on";

        /* (non-Javadoc)
         * @see com.vaadin.event.dd.DropHandler#drop(com.vaadin.event.dd.DragAndDropEvent)
         */
        @Override
        public void drop(DragAndDropEvent event) {            
            
            //If not licensed do nothing
            if(!getState().licensed) {
                return;
            }
            
            //Grab the dropped object
            Object dropObject = getDataFromDragAndDropEvent(event);
            
            //Determine if a user item was dropped on
            Double itemId = (Double)event.getTargetDetails().getData(DROP_ITEM_DETAIL_ID);
            UserComponentItem item = null;
            
            if(itemId != null) {
                item = userButtonItemMap.get(itemId);
                if(item == null) {
                    item = userAccountPopupItemMap.get(itemId);
                }
            }
            
            //Fire event
            fireEvent(new UserComponent.ObjectDropEvent(event, UserComponent.this, dropObject, item));
        }

        /* (non-Javadoc)
         * @see com.vaadin.event.dd.DropHandler#getAcceptCriterion()
         */
        @Override
        public AcceptCriterion getAcceptCriterion() {
            return new ClassTypeCriterion();
        }
        
    }
    
    /**
     * TypeCriterion that uses the acceptedDropClasses to verify drop acceptance.
     */
    public class ClassTypeCriterion extends ServerSideCriterion {

        /** The Constant serialVersionUID. */
        private static final long serialVersionUID = 1L;
                
        /* (non-Javadoc)
         * @see com.vaadin.event.dd.acceptcriteria.AcceptCriterion#accept(com.vaadin.event.dd.DragAndDropEvent)
         */
        @Override
        public boolean accept(DragAndDropEvent dragEvent) {
                
            //Check if the destination is the same as the source if so don't accept
            if(dragEvent.getTransferable().getSourceComponent().equals(UserComponent.this)) {
                return false;
            }
                            
            //If accept classes are set check if any match if so accept
            if(acceptedDropClasses != null) {
                Object data = null;
                try {
                    data = getDataFromDragAndDropEvent(dragEvent);
                } catch (Exception e) {
                    return false;
                }

                for(Class<?> classType : acceptedDropClasses) {
                    if(classType.isAssignableFrom(data.getClass())) {
                        return true;
                    }
                }
            }
            
            return false;
        }
    }

}
